-----------------------------------------
-- ALU
-----------------------------------------

---- Description --------------------------------------------------
-- ALU Function
---- Arithmetic
---- Compare
---- Logic
---- Shift
--
-- cf."߃Zbg߂܂傤"
-- OP is operation code (op2)
-- SRC1 := R2, SRC2 := R3 or Imm(sign extended);
-- R1 := DST;
-------------------------------------------------------------------

---- Notes --------------------------------------------------------
-- 2006.12.31
--   ƏĂ݂܂BALU߂͑SĎB
--   l̓rbggSRC2ɕ荞܂\B
--   gݍ킹HȂ̂ŃNbN͂܂B
--   shiftǂȉHɂȂ邩\złȂ̂|łB
--------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_signed.all;

entity ALU is
    port( OP : in std_logic_vector(3 downto 0);          --- opcode
          SRC1,SRC2 : in std_logic_vector(31 downto 0);  --- R2 -> SRC1, R3 -> SRC2
          DST : out std_logic_vector(31 downto 0));      --- output
end ALU;


architecture RTL of ALU is
    signal MULOUT : std_logic_vector(63 downto 0);
    signal SINTL,SINTRL,SINTRA : Integer;
    constant ALL0 : std_logic_vector(31 downto 0) := (others => '0');
    constant ALL1 : std_logic_vector(31 downto 0) := (others => '1');
    
begin
    process(OP)
    begin
       case OP is
           -- arith
           when "0000" => -- add,addi
               DST <= SRC1 + SRC2;
           when "0001" => -- sub
               DST <= SRC1 - SRC2;
           when "0010" => -- mul,muli
               MULOUT <= SRC1 * SRC2;
               DST <= MULOUT(31 downto 0);
           when "0011" => -- mulh
               DST <= MULOUT(63 downto 32);
           when "0100" => -- div
               DST <= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(SRC1) / CONV_INTEGER(SRC2),32);
           -- compare
           when "0101" => -- eq,eqi
               if SRC1 = SRC2 then
                   DST <= (others => '1');
               else
                   DST <= (others => '0');
               end if;
           when "0110" => -- lt,lti
               if SRC1 < SRC2 then
                   DST <= (others => '1');
               else
                   DST <= (others => '0');
               end if;
           when "0111" => -- gti
               if SRC1 > SRC2 then
                   DST <= (others => '1');
               else
                   DST <= (others => '0');
               end if;
           -- logic
           when "1000" => -- and, andi
               DST <= SRC1 and SRC2;
           when "1001" => -- or,ori
               DST <= SRC1 or SRC2;
           when "1010" => -- xor,xori
               DST <= SRC1 xor SRC2;
           when "1011" => -- not
               DST <= not SRC1;
--           when "10001" => -- andi
--            DST <= SRC1 and ("000000000000000000000" & SRC2(10 downto 0));
--        when "10011" => -- ori
--            DST <= SRC1 or ("000000000000000000000" & SRC2(10 downto 0));
--        when "10101" => -- xori
--            DST <= SRC1 xor ("000000000000000000000" & SRC2(10 downto 0));
        -- shift
           when "1100" => -- sl,sli
               SINTL <= CONV_INTEGER(('0' & SRC2(5 downto 0)));
               if SINTL >= 32 then
                   DST <= (others => '0');
               else
                   DST <= SRC1(32-SINTL downto 0) & ALL0(SINTL-1 downto 0);
               end if;
           when "1101" => -- srl,srli
               SINTRL <= CONV_INTEGER(('0' & SRC2(5 downto 0)));
               if SINTRL >= 32 then
                   DST <= (others => '0');
               else
                   DST <= ALL0(SINTRL-1 downto 0) & SRC1(31 downto SINTRL);
               end if;
           when "1110" => -- sra,srai
               SINTRA <= CONV_INTEGER(('0' & SRC2(5 downto 0)));
               if SINTRA >= 32 then
                   DST <= (others => SRC1(31));
               else
                   if SRC1(31) = '1' then
                       DST <= ALL1(SINTRA-1 downto 0) & SRC1(31 downto SINTRA);
                   else
                       DST <= ALL0(SINTRA-1 downto 0) & SRC1(31 downto SINTRA);
                   end if;
               end if;
           when others => DST <= (others => 'X');
       end case;
   end process;
   
end RTL;
