----------------------------------------
-- Instruction Execute & Write Phase
----------------------------------------

-------------------------------------------------------------------------------
-- Notes
--  2007.01.27
---- Ə܂BSRAM̃^C~Ô͂B
-- 2007.02.11
---- USBgݍ݁BH삵ĂA܂傤B
---- ȂɂႢĂƍ̂łˁB
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_signed.all;

entity EXW is
    port(CLK : in std_logic;     -- system clock
        RESET : in std_logic;    -- reset
        -- Mr.SRAM
        SRAM_RW : out std_logic;                        --sram rw
        SRAM_ADDR : out std_logic_vector(19 downto 0);  -- sram addr
        -- SRAM
        SRAM_WD : out std_logic_vector(31 downto 0); -- to sram, write data
        SRAM_RD : in std_logic_vector(31 downto 0); -- from Sram, read data
        -- IFDR
        PC : in std_logic_vector(19 downto 0);  -- instruction address(PC)
        EXPC : out std_logic_vector(19 downto 0);   -- External PC
        BUSY : out std_logic;                       -- EXW busy signal
        VALID : in std_logic;                    -- FDR valid
        OP1 : in std_logic_vector(1 downto 0);   -- opcode1
        OP2 : in std_logic_vector(3 downto 0);   -- opcode2
        REGAD : in std_logic_vector(6 downto 0);  -- desg reg addr/imm7,disp7
        IMMSEL : in std_logic;                    -- if IMM then 1 else 0
        IMM : in std_logic_vector(31 downto 0);   -- imm
        -- register file
--        REGSEL : out std_logic;                  -- select signal 1->EXW 0->FDR
        REGWE : out std_logic;                    -- regfile write enable(0->read)
        REGAD_D : out std_logic_vector(6 downto 0); -- dest reg addr
        REG_D : out std_logic_vector(31 downto 0); -- regfile write data
        REG_S : in std_logic_vector(31 downto 0); -- R2 src data
        REG_T : in std_logic_vector(31 downto 0); -- R3/R1 src data
        -- usb
        USB_RMODE : out std_logic;
        USB_WMODE : out std_logic;
        USB_RDATA : in std_logic_vector(7 downto 0);
        USB_WDATA : out std_logic_vector(7 downto 0);
        USB_VALID : in std_logic;
        USB_BUSY  : in std_logic

);
end EXW;

architecture ARC of EXW is

    component ALU is
        port( OP : in std_logic_vector(3 downto 0);          --- opcode
          SRC1,SRC2 : in std_logic_vector(31 downto 0);  --- R2 -> SRC1, R3 -> SRC2
          DST : out std_logic_vector(31 downto 0));
    end component;

    constant ALL0 : std_logic_vector(31 downto 0) := (others => '0'); -- all 0
    signal STATE : std_logic_vector(3 downto 0);   -- state
    signal REGB : std_logic_vector(6 downto 0);   -- reg read/write addr
    signal RETAD : std_logic_vector(19 downto 0);  -- return addr(->R127)
    signal EXPCB : std_logic_vector(19 downto 0);  -- exPC buffer
    signal SADB : std_logic_vector(19 downto 0); -- sram addr buffer
    signal SWD : std_logic_vector(31 downto 0);
    signal SA : std_logic_vector(19 downto 0);
    signal SMODE : std_logic;
    -- ALU
    signal ALUOP : std_logic_vector(3 downto 0);
    signal ALUS1 : std_logic_vector(31 downto 0);
    signal ALUS2 : std_logic_vector(31 downto 0);
    signal ALUD : std_logic_vector(31 downto 0);
    -- exdevice
    signal EXSTATE : std_logic_vector(3 downto 0);   -- for exdevice state machine
    signal USB_WDBUF : std_logic_vector(23 downto 0); -- usb write data buffer
    signal USBWBUF1,USBWBUF2,USBWBUF3 : std_logic_vector(7 downto 0); -- usb read data buffer


begin

    ALU_INST : ALU
        port map(OP => ALUOP,
                SRC1 => ALUS1,
                SRC2 => ALUS2,
                DST  => ALUD);

    EXPC <= EXPCB;
    -- sram
    SRAM_ADDR <= SA;
    SRAM_WD <= SWD;
    SRAM_RW <= SMODE;

    process(CLK,RESET)
    begin
        if RESET = '1' then
            EXPCB <= (others => '0');
            STATE <= "1110";
            REGB <= (others =>'Z');
            RETAD <= (others => 'Z');
            -- r0 is 0 register
            REGAD_D <= (others => '0');
            REG_D <= (others => '0');
            REGWE <= '1';

        elsif CLK'event and CLK = '1' then
            case STATE is
                when "0000"=>
--                    REGSEL <= '0';
                    if VALID = '1' then
                        BUSY <= '1';
                       case OP1 is
                            when "00" => -- ALU
                                ALUOP <= OP2;
                                REGB <= REGAD;
                                ALUS1 <= REG_S;
                                if IMMSEL = '1' then
                                    ALUS2 <= IMM;
                                else
                                    ALUS2 <= REG_T;
                                end if;
                                STATE <= "0001";
                            when "01" => -- FPU
                            when "10" => -- Memory
                                case OP2(1) is
                                    when '0' =>
                                        case OP2(3 downto 2) is
                                            when "00" => -- imm
                                                REGWE <= '1';
                                                REGAD_D <= REGAD;
                                                REG_D <= IMM(20 downto 0) & "00000000000";
                                                STATE <= "0101";
                                            when "11" => -- store
                                                if (IMMSEL xor OP2(0)) = '1' then
                                                   STATE <= "0110";
                                                   REGWE <= '0';
                                                   REGAD_D <= REGAD;
                                                   SADB <= REG_S(19 downto 0) + REG_T(19 downto 0);
                                                else
                                                   SMODE <= '0';
                                                   if IMMSEL = '1' then
                                                       SA <= REG_S(19 downto 0) + IMM(19 downto 0);
                                                   else
                                                       SA <= REG_S(19 downto 0);
                                                   end if;
                                                   SWD <= REG_T;
                                                   STATE <= "1110";
                                                end if;
                                            when "10" => -- load
                                                SMODE <= '0';
                                                case OP2(0) is
                                                    when '0'=>
                                                        case IMMSEL is
                                                            when '0' =>
                                                                SA <= REG_S(19 downto 0);    
                                                            when others =>
                                                                SA <= REG_S(19 downto 0);
                                                        end case;
                                                    when '1' =>
                                                       case IMMSEL is
                                                          when '0' =>
                                                             SA <= REG_S(19 downto 0) + REG_T(19 downto 0);
                                                          when '1' =>
                                                              SA <= REG_S(19 downto 0) + IMM(19 downto 0);
                                                          when others =>
                                                               SA <= REG_S(19 downto 0);
                                                    end case;
                                                    when others =>
                                                        SA <= REG_S(19 downto 0);
                                                end case;
--                                                case (OP2(0) & IMMSEL) is
  --                                                  when "00" =>
    --                                                   SA <= REG_S(19 downto 0);
      --                                              when "10" =>
        --                                               SA <= REG_S(19 downto 0) + REG_T(19 downto 0);
          --                                          when "11" =>
            --                                           SA <= REG_S(19 downto 0) + IMM(19 downto 0);
              --                                      when others=>
                --                                       SA <= REG_S(19 downto 0);
                  --                              end case;
                                                STATE <= "0111";
                                                REGB <= REGAD;
                                            when others =>
                                                STATE <= "1110";
                                        end case;
                                    when '1' => -- usb or rs232c
                                         case OP2(3 downto 2) is
                                            when "11" => -- to usb "out"
                                                  USB_WDATA <= REG_S(31 downto 24);
                                                  USB_WDBUF <= REG_S(23 downto 0);
                                                  USB_RMODE <= '0';
                                                  STATE <= "1010";
                                                  EXSTATE <= "0000";
                                            when "10" => -- from usb "in"
                                                  REGB <= REGAD;
                                                  USB_WMODE <= '0';
                                                  STATE <="1010";
                                                  EXSTATE <= "1000";
                                            when others =>
                                                STATE <= "1110";
                                         end case;
                                    when others =>
                                        STATE <= "1110";
                                end case;
                            when "11" => -- Jump
                                if OP2(3 downto 2) = "00" then
                                    case OP2(1 downto 0) is
                                        when "00" => -- jump when R1 = 0
                                            if REG_T = ALL0 then
                                                --EXPCB <= PC + IMM(19 downto 0);
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                            else
                                               EXPCB <= PC;
                                                --EXPCB <= PC + 1;
                                            end if;
                                        when "01" => -- jump when R1 != 0
                                            if REG_T /= ALL0 then
                                                --EXPCB <= PC + IMM(19 downto 0);
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                            else
                                               EXPCB <= PC;
                                                --EXPCB <= PC + 1;
                                            end if;
                                        when "10" => -- jump when R1 > 0
                                            if REG_T > ALL0 then
                                                --EXPCB <= PC + IMM(19 downto 0);
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                            else
                                               EXPCB <= PC;
                                               --EXPCB <= PC + 1;
                                            end if;
                                        when "11" => -- jump when R1 >= 0
                                            if REG_T >= ALL0 then
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                                --EXPCB <= PC + IMM(19 downto 0);
                                            else
                                               EXPCB <= PC;
                                               --EXPCB <= PC + 1;
                                            end if;
                                        when others =>
                                            EXPCB <= PC;
                                            --EXPCB <= PC + 1;
                                    end case;
                                    STATE <= "0011";
                                else
                                    case OP2 is
                                        when "1100" =>
                                            --EXPCB <= PC + (REGAD(0) & IMM(18 downto 0));
                                            EXPCB <= PC + (REGAD(0) & IMM(18 downto 0))-1;
                                            STATE <= "0011";
                                        when "1101" =>
                                            --EXPCB <= PC + (REGAD(0) & IMM(18 downto 0));
                                            EXPCB <= PC + (REGAD(0) & IMM(18 downto 0))-1;
                                            REGB <= "1111111";
                                            RETAD <= PC;
                                            --RETAD <= PC + 1;
                                            STATE <= "0100";
                                        when "0100" =>
                                            EXPCB <= REG_S(19 downto 0)-1;
                                            --EXPCB <= REG_S(19 downto 0);
                                            STATE <= "0011";
                                        when "0101" =>
                                            --EXPCB <= REG_S(19 downto 0);
                                            EXPCB <= REG_S(19 downto 0);
                                            REGB <= "1111111";
                                            RETAD <= PC;
                                            --RETAD <= PC + 1;
                                            STATE <= "0100";
                                        when "0110" =>
                                            EXPCB <= REG_S(19 downto 0)-1;
                                            --EXPCB <= REG_S(19 downto 0);
                                            STATE <= "0011";
                                        when "0111" =>
                                            STATE <= "1111";
                                        when others =>
                                            EXPCB <= PC;
                                            --EXPCB <= PC + 1;
                                            STATE <= "0011";
                                    end case;
                                end if;
                            when others =>
                                STATE <= "1110";
                        end case;
                    else
                        STATE <= "0000";
                    end if;
                when "0001"=> -- ALU 2
                    REGWE <= '1';
                    REGAD_D <= REGB;
                    REG_D <= ALUD;
                    STATE <= "1110";
                when "0010" => -- FPU 2
                    
                when "0011"=> -- Jump 2-1
                    STATE <= "1110";
                when "0100"=> -- Jump 2-2
                    REGWE <= '1';
                    REGAD_D <= REGB;
                    REG_D <= "000000000000" & RETAD;
                    STATE <= "1110";
                when "0101" =>  -- Memory IMM 2
                    STATE <= "1110";
                when "0110" =>  -- Memory Store 2
                    SMODE <= '0';
                    SA <= SADB;
                    SWD <= REG_T;
                    STATE <= "1110";
                when "0111" =>  -- Memory Load 2
                    STATE <= STATE + 1;
                when "1000" =>  -- Memory Load 3
                    STATE <= STATE + 1;
                when "1001" => -- Memory Load 4
                    REGWE <= '1';
                    REGAD_D <= REGB;
                    REG_D <= SRAM_RD;
                    STATE <= "1110";
                when "1010" => -- exdevice
                    REGB <= REGB;
                    case EXSTATE is
                        when "0000" =>  -- out
                            if USB_BUSY = '0' then
                                USB_WMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "0000";
                            end if;
                            STATE <= "1010";
                        when "0001" =>     -- USB write 2-1
                            if USB_VALID = '1' then
                                USB_WMODE <= '0';
                                USB_WDATA <= USB_WDBUF(23 downto 16);
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "0001";
                            end if;
                            STATE <= "1010";
                        when "0010" =>     -- USB write 2-2
                            if USB_BUSY = '0' then
                                USB_WMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "0010";
                            end if;
                            STATE <= "1010";
                        when "0011" =>     -- USB write 3-1
                            if USB_VALID = '1' then
                                USB_WMODE <= '0';
                                USB_WDATA <= USB_WDBUF(15 downto 8);
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "0011";
                            end if;
                            STATE <= "1010";
                        when "0100" =>     -- USB write 3-2
                            if USB_BUSY = '0' then
                                USB_WMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "0100";
                            end if;
                            STATE <= "0011";
                        when "0101" =>     -- USB write 4-1
                            if USB_VALID = '1' then
                                USB_WMODE <= '0';
                                USB_WDATA <= USB_WDBUF(7 downto 0);
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "0101";
                            end if;
                            STATE <= "1010";
                        when "0110" =>     -- USB write 4-2
                            if USB_BUSY = '0' then
                                USB_WMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "0110";
                            end if;
                            STATE <= "1010";
                        when "0111" =>      -- USB write end
                            if USB_VALID = '1' then    
                                USB_WMODE <= '0';
                                STATE <= "1110";
                                EXSTATE <= "0000";
                            else
                                EXSTATE <= "0111";
                                STATE <= "1010";
                            end if;
                        when "1000" =>  -- in
                            if USB_BUSY = '0' then  --- USB read 1-1
                                USB_RMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1000";
                            end if;
                            STATE <= "1010";
                        when "1001" =>
                            if USB_VALID = '1' then --- USB read 1-2
                                USB_RMODE <= '0';
                                USBWBUF1 <= USB_RDATA;
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1001";
                            end if;
                            STATE <= "1010";
                        when "1010" =>
                            if USB_BUSY = '0' then  --- USB read 2-1
                                USB_RMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1010";
                            end if;
                            STATE <= "1010";
                        when "1011" =>
                            if USB_VALID = '1' then --- USB read 2-2
                                USB_RMODE <= '0';
                                USBWBUF2 <= USB_RDATA;
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1011";
                            end if;
                            STATE <= "1010";
                        when "1100" =>
                            if USB_BUSY = '0' then  --- USB read 3-1
                                USB_RMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1100";
                            end if;
                            STATE <= "1010";
                        when "1101" =>
                            if USB_VALID = '1' then --- USB read 3-2
                                USB_RMODE <= '0';
                                USBWBUF3 <= USB_RDATA;
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1101";
                            end if;
                            STATE <= "1010";
                        when "1110" =>
                            if USB_BUSY = '0' then  --- USB read 4-1
                                USB_RMODE <= '1';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1110";
                            end if;
                            STATE <= "1010";
                        when "1111" =>
                            if USB_VALID = '1' then --- USB read 4-2
                                USB_RMODE <= '0';
                                REGWE <= '1';
                                REGAD_D <= REGB;
                                REG_D <= USBWBUF1 & USBWBUF2 & USBWBUF3 & USB_RDATA;
                                EXSTATE <= "1000";
                                STATE <= "1110";
                            else
                                EXSTATE <= "1111";
                                STATE <= "1010";
                            end if;
                        when others =>
                            STATE <= "1110";
                            EXSTATE <= "0000";
                            USB_WMODE <= '0';
                            USB_RMODE <= '0';
                    end case;
                when "1110" => -- next return
                    SMODE <='1';
                    REGWE <= '0';
                    BUSY <= '0';
                    STATE <= "0000";
                when "1111" => -- HALT
                    STATE <= "1111";
                when others =>
                    STATE <= "0000";
            end case;
        end if;
    end process;

end ARC;

